@@ -68,6 +68,7 @@ to the user.
 """
 
 import os, tempfile, shutil, subprocess, re
+from datetime import datetime
 
 from StringIO import StringIO
 
@@ -175,7 +176,10 @@ class Context():
         we can optionnally watch for a confirmation pattern on the
         statusfd.
         """
-        proc = subprocess.Popen(self.build_command(command), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        proc = subprocess.Popen(self.build_command(command),  # nosec
+                                stdin=subprocess.PIPE,
+                                stdout=subprocess.PIPE,
+                                stderr=subprocess.PIPE)
         (self.stdout, self.stderr) = proc.communicate(stdin)
         self.returncode = proc.returncode
         if self.debug:
@@ -224,8 +228,17 @@ class Context():
 
         this therefore looks only at the next line, but may also hang
         like seek_pattern()
+
+        if the beginning of the line matches a pattern which is
+        being ignored, it will skip it and look at the next line
         """
         line = fd.readline()
+        ignored = ('[GNUPG:] KEYEXPIRED', '[GNUPG:] SIGEXPIRED', 'gpg: ')
+
+        while line and line.startswith(ignored):
+            if self.debug: print >>self.debug, "SKIPPED:", line,
+            line = fd.readline()
+
         match = re.search(pattern, line)
 
         if self.debug:
@@ -251,7 +264,7 @@ class Context():
         if self.debug:
             print >>self.debug, "WROTE:", message
         print >>fd, message
-    
+
     def version(self):
         """return the version of the GPG binary"""
         self.call_command(['version'])
@@ -351,37 +364,23 @@ class Keyring():
         self.context.call_command(['recv-keys', fpr])
         return self.context.returncode == 0
 
-    def get_keys(self, pattern = None, secret = False, public = True):
+    def get_keys(self, pattern=None, secret=False, public=True, keys=None):
         """load keys matching a specific patterns
 
         this uses the (rather poor) list-keys API to load keys
         information
         """
-        keys = {}
-        if public:
-            command = ['list-keys']
-            if pattern: command += [pattern]
-            self.context.call_command(command)
-            if self.context.returncode == 0:
-                # discard trustdb data, first line of output
-                self.context.stdout = "\n".join(self.context.stdout.split("\n")[1:])
-                for keydata in self.context.stdout.split("pub:"):
-                    if not keydata: continue
-                    keydata = "pub:" + keydata
-                    key = OpenPGPkey(keydata)
-                    keys[key.fpr] = key
-            elif self.context.returncode == 2:
-                return None
-            else:
-                raise GpgProtocolError(self.context.returncode, _('unexpected GPG exit code in list-keys: %d') % self.context.returncode)
+        if keys is None:
+            keys = {}
         if secret:
             command = ['list-secret-keys']
             if pattern: command += [pattern]
             self.context.call_command(command)
             if self.context.returncode == 0:
-                for keydata in self.context.stdout.split("sec::"):
+                needle = 'sec:'
+                for keydata in self.context.stdout.split(needle):
                     if not keydata: continue
-                    keydata = "sec::" + keydata
+                    keydata = needle + keydata
                     key = OpenPGPkey(keydata)
                     # check if we already have that key, in which case we
                     # add to it instead of adding a new key
@@ -394,6 +393,35 @@ class Keyring():
                 return None
             else:
                 raise GpgProtocolError(self.context.returncode, _('unexpected GPG exit code in list-keys: %d') % self.context.returncode)
+        if public:
+            command = ['list-keys']
+            if pattern: command += [pattern]
+            self.context.call_command(command)
+            if self.context.returncode == 0:
+                # discard trustdb data, first line of output
+                self.context.stdout = "\n".join(self.context.stdout.split("\n")[1:])
+                for keydata in self.context.stdout.split("pub:"):
+                    if not keydata: continue
+                    keydata = "pub:" + keydata
+                    key = OpenPGPkey(keydata)
+                    # check if we already have that key, in which case we
+                    # add to it instead of adding a new key
+                    if key.fpr in keys:
+                        keys[key.fpr].parse_gpg_list(self.context.stdout)
+                        del key
+                    else:
+                        keys[key.fpr] = key
+            elif self.context.returncode == 2:
+                return None
+            else:
+                raise GpgProtocolError(self.context.returncode, _('unexpected GPG exit code in list-keys: %d') % self.context.returncode)
+        if secret:
+            # workaround for bug in GPG < 2.1 where revocation
+            # information is not correctly populated in
+            # --list-secret-keys --with-colons
+            for fpr, key in keys.items():
+                if key.get_trust() == 'empty':
+                    self.get_keys(fpr, secret=False, public=True, keys=keys)
         return keys
 
     def encrypt_data(self, data, recipient):
@@ -419,8 +447,13 @@ class Keyring():
             raise GpgRuntimeError(self.context.returncode, _('decryption failed: %s') % self.context.stderr.split("\n")[-2])
 
     def del_uid(self, fingerprint, pattern):
-        if self.context.debug: print >>self.context.debug, 'command:', self.context.build_command(['edit-key', fingerprint])
-        proc = subprocess.Popen(self.context.build_command(['edit-key', fingerprint]), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        command = self.context.build_command(['edit-key', fingerprint])
+        if self.context.debug:
+            print >>self.context.debug, 'command:', command
+        proc = subprocess.Popen(command,  # nosec
+                                stdin=subprocess.PIPE,
+                                stdout=subprocess.PIPE,
+                                stderr=subprocess.PIPE)
         # start copy-paste from sign_key()
         self.context.expect(proc.stderr, 'GET_LINE keyedit.prompt')
         while True:
@@ -460,8 +493,14 @@ class Keyring():
         # we iterate over the keys matching the provided
         # keyid, but we should really load those uids from the
         # output of --sign-key
-        if self.context.debug: print >>self.context.debug, 'command:', self.context.build_command([['sign-key', 'lsign-key'][local], pattern])
-        proc = subprocess.Popen(self.context.build_command([['sign-key', 'lsign-key'][local], pattern]), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        command = self.context.build_command([['sign-key',
+                                               'lsign-key'][local], pattern])
+        if self.context.debug:
+            print >>self.context.debug, 'command:', command
+        proc = subprocess.Popen(command,  # nosec
+                                stdin=subprocess.PIPE,
+                                stdout=subprocess.PIPE,
+                                stderr=subprocess.PIPE)
 
         # if there are multiple uids to sign, we'll get this point, and a whole other interface
         try:
@@ -469,15 +508,6 @@ class Keyring():
         except GpgProtocolError as e:
             if 'sign_uid.okay' in str(e):
                 multiuid = False
-            elif '[GNUPG:]' not in str(e):
-                # this is not a protocol message, try again but skipping now
-                # this was necessary in order to sign Zack's key, as it was spewing:
-                # gpg: moving a key signature to the correct place
-                # instead of a [GNUGPG:] message
-                try:
-                    multiuid = self.context.seek(proc.stderr, 'GET_BOOL keyedit.sign_all.okay')
-                except GpgProtocolError as e:
-                    raise GpgRuntimeError(self.context.returncode, _('cannot select uid for signing: %s') % e.found().decode('utf-8'))
             else:
                 raise GpgRuntimeError(self.context.returncode, _('cannot select uid for signing: %s') % e.found().decode('utf-8'))
         if multiuid:
@@ -488,7 +518,7 @@ class Keyring():
                 try:
                     self.context.expect(proc.stderr, 'GET_BOOL sign_uid.okay')
                 except GpgProtocolError as e:
-                    if 'sign_uid.dupe_okay' in str(e):
+                    if 'ALREADY_SIGNED' in str(e) or 'sign_uid.dupe_okay' in str(e):
                         raise GpgRuntimeError(self.context.returncode, _('you already signed that key'))
                     else:
                         # propagate gpg error message up
@@ -566,13 +596,34 @@ class OpenPGPkey():
     Some of this datastructure is taken verbatim from GPGME.
     """
 
-    # the key has a revocation certificate
-    # @todo - not implemented
-    revoked = False
+    @property
+    def revoked(self):
+        '''Returns whether GnuPG thinks the key has been revoked
 
-    # the expiry date is set and it is passed
-    # @todo - not implemented
-    expired = False
+        This is the second field of the result of the --list-key --with-colons
+        call.  Note that this information is only present on public keys,
+        i.e. not on secret keys.
+
+        Returns None if it cannot be determined whether this key has
+        been revoked.'''
+        if self.trust == '':
+            # We cannot determine whether this key has been revoked.
+            # Locate the public key and try again.
+            is_revoked = None
+        elif self.trust == 'r':
+            is_revoked = True
+        else:
+            is_revoked = False
+            
+        return is_revoked
+
+    @property
+    def expired(self):
+        if not self.expiry:
+            ret = False
+        else:
+            ret = datetime.now() > self.expiry
+        return ret
 
     # the key has been disabled
     # @todo - not implemented
@@ -611,7 +662,24 @@ class OpenPGPkey():
 
     # This is the expiration timestamp of the subkey, or 0 if the
     # subkey does not expire.
-    expiry = 0
+    _expiry = 0
+    
+    @property
+    def expiry(self):
+        """Returns a datetime from the _expiry field or None if
+        the key does not expire"""
+        if self._expiry:
+            expiry = int(self._expiry)
+            if expiry > 0:
+                result = datetime.fromtimestamp(expiry)
+            else:
+                result = None
+        elif self._expiry == '':
+                result = None
+        else:
+            raise ValueError("Expiry is %r but we expected an int or string",
+                             self._expiry)
+        return result
 
     # single-character trust status, see trust_map below for parsing
     trust = None
@@ -642,6 +710,9 @@ class OpenPGPkey():
                                   # This often means that the secret
                                   # key is available, but any key may
                                   # be marked as ultimately trusted.
+                 '': 'empty',  # trust value was not communicated by gpg
+                               # this is a bug in gpg 2.1: secret keys do
+                               # not have trust information attached
                  }
 
     def __init__(self, data=None):
@@ -657,7 +728,7 @@ class OpenPGPkey():
 
     def keyid(self, l=8):
         if self.fpr is None:
-            assert(self._keyid is not None)
+            assert(self._keyid is not None)  # nosec
             return self._keyid[-l:]
         return self.fpr[-l:]
 
@@ -673,35 +744,31 @@ class OpenPGPkey():
             #print >>sys.stderr, "\n"
             rectype = record[0]
             if rectype == 'tru':
-                (rectype, trust, selflen, algo, keyid, creation, expiry, serial) = record
+                (rectype, trust, selflen, algo, keyid, creation, expiry, serial) = record[:8]
             elif rectype == 'fpr':
                 if not self.fpr:
                     self.fpr = record[9]
             elif rectype == 'pub':
-                (null, self.trust, self.length, self.algo, keyid, self.creation, self.expiry, serial, trust, uid, sigclass, purpose, smime) = record
+                (null, self.trust, self.length, self.algo, keyid, self.creation, self._expiry, serial, trust, uid, sigclass, purpose, smime) = record[:13]
                 for p in self.purpose:
                     self.purpose[p] = p[0].lower() in purpose.lower()
-                if self.trust == '':
-                    self.trust = '-'
             elif rectype == 'uid':
-                (rectype, trust, null  , null, null, creation, expiry, uidhash, null, uid, null) = record
+                (rectype, trust, null, null, null, creation, expiry, uidhash, null, uid) = record[:10]
                 uid = OpenPGPuid(uid, trust, creation, expiry, uidhash)
                 self.uids[uidhash] = uid
                 uidslist.append(uid)
             elif rectype == 'sub':
                 subkey = OpenPGPkey()
-                (rectype, trust, subkey.length, subkey.algo, subkey._keyid, subkey.creation, subkey.expiry, serial, trust, uid, sigclass, purpose, smime) = record
+                (rectype, trust, subkey.length, subkey.algo, subkey._keyid, subkey.creation, subkey._expiry, serial, trust, uid, sigclass, purpose, smime) = record[:13]
                 for p in subkey.purpose:
                     subkey.purpose[p] = p[0].lower() in purpose.lower()
                 self.subkeys[subkey._keyid] = subkey
             elif rectype == 'sec':
-                (null, self.trust, self.length, self.algo, keyid, self.creation, self.expiry, serial, trust, uid, sigclass, purpose, smime, wtf, wtf, wtf) = record
+                (null, self.trust, self.length, self.algo, keyid, self.creation, self._expiry, serial, trust, uid, sigclass, purpose, smime) = record[:13]
                 self.secret = True
-                if self.trust == '':
-                    self.trust = '-'
             elif rectype == 'ssb':
                 subkey = OpenPGPkey()
-                (rectype, trust, subkey.length, subkey.algo, subkey._keyid, subkey.creation, subkey.expiry, serial, trust, uid, sigclass, purpose, smime, wtf, wtf, wtf) = record
+                (rectype, trust, subkey.length, subkey.algo, subkey._keyid, subkey.creation, subkey._expiry, serial, trust, uid, sigclass, purpose, smime) = record[:13]
                 if subkey._keyid in self.subkeys:
                     # XXX: nothing else to add here?
                     self.subkeys[subkey._keyid].secret = True
@@ -714,25 +781,36 @@ class OpenPGPkey():
             elif rectype == '':
                 pass
             else:
-                raise NotImplementedError(_("record type '%s' not implemented") % rectype)
-        if uidslist: self.uidslist = uidslist
+                # XXX: need to log this
+                #raise NotImplementedError(_("record type '%s' not implemented") % rectype)
+                pass  # ignore unknown records for forward-compatibilty
+        self.uidslist = uidslist
+
 
     def __str__(self):
         ret = u'pub  [%s] %sR/' % (self.get_trust(), self.length)
         ret += self.keyid(8) + u" " + self.creation
-        if self.expiry: ret += u' [expiry: ' + self.expiry + ']'
+        if self._expiry: ret += u' [expiry: ' + str(self.expiry) + ']'
         ret += u"\n"
         ret += u'    Fingerprint = ' + self.format_fpr() + "\n"
         i = 1
         for uid in self.uidslist:
-            ret += u"uid %d      [%s] %s\n" % (i, uid.get_trust(), uid.uid.decode('utf-8'))
+            ret += u"uid %d      [%s] %s\n" % (i, uid.get_trust(), uid.uid.decode('utf-8', 'replace'))
             i += 1
         for subkey in self.subkeys.values():
             ret += u"sub   " + subkey.length + u"R/" + subkey.keyid(8) + u" " + subkey.creation
-            if subkey.expiry: ret += u' [expiry: ' + subkey.expiry + "]"
+            if subkey._expiry: ret += u' [expiry: ' + str(subkey.expiry) + "]"
             ret += u"\n"
         return ret
 
+
+    def __repr__(self):
+        s = '<OpenPGPKey(%s UIDs:%d)>'
+        s %= (self.fpr,  len(self.uidslist))
+        
+        return s
+
+
     def format_fpr(self):
         """display a clean version of the fingerprint
 
@@ -753,15 +831,38 @@ class OpenPGPuid():
     def __init__(self, uid, trust, creation = 0, expire = None, uidhash = ''):
         self.uid = uid
         self.trust = trust
-        if self.trust == '':
-            self.trust = '-'
         self.creation = creation
         self.expire = expire
         self.uidhash = uidhash
 
+    @property
+    def revoked(self):
+        '''Whether this UID has been revoked
+
+        Note that, due to GnuPG not exporting that information
+        for secret keys, UIDs of secret keys do not carry that
+        information.
+
+        Return None if it cannot be determined whether this UID
+        has been revoked.  Try again with the public key.'''
+        if self.trust == '':
+            is_revoked = None
+        elif self.trust == 'r':
+            is_revoked = True
+        else:
+            is_revoked = False
+        return is_revoked
+
     def get_trust(self):
         return OpenPGPkey.trust_map[self.trust]
 
+    def __repr__(self):
+        s = '<OpenPGPuid(%s) [%s:%s:%s]>' 
+        s %= (self.uid, self.trust, self.creation, self.expire)
+
+        return s
+
+
 class GpgProtocolError(IOError):
     """simple exception raised when we have trouble talking with GPG
 
