@@ -17,7 +17,7 @@
 
 from monkeysign import __version__
 # gpg interface
-from monkeysign.gpg import Keyring, TempKeyring, GpgRuntimeError
+from monkeysign.gpg import Keyring, TempKeyring, GpgRuntimeError, Context
 import monkeysign.translation
 
 # mail functions
@@ -30,15 +30,228 @@ from email.header import Header
 from email.utils import parseaddr, formataddr
 from email import Charset
 import smtplib
-import subprocess
+import subprocess  # nosec
 
 # system libraries
-import optparse
+import argparse
+import errno
+import getpass
+import os
+import platform
 import sys
 import re
-import os
+import shlex
+try:
+    from shlex import quote as cmd_quote
+except ImportError:
+    from pipes import quote as cmd_quote
 import shutil
 import socket
+import tempfile
+
+
+class MonkeysignArgumentParser(argparse.ArgumentParser):
+    # config files, in priority order (first is higher)
+    configs = ['~/.config/monkeysign.conf', '/etc/monkeysign.conf']
+
+    # standard header that needs to be present in the config file for
+    # us to rewrite it
+    header = """# this file was created by monkeysign
+# comments will be lost on next write"""
+
+    # those parameters will not be saved to the config file
+    ephemeral = ['save', 'version', 'help', 'pattern', 'to']
+
+    def __init__(self,
+                 prog=None, usage=None, description=None, epilog=None,
+                 parents=[], formatter_class=argparse.HelpFormatter,
+                 prefix_chars='-', fromfile_prefix_chars=None,
+                 argument_default=None, conflict_handler='error',
+                 add_help=True, allow_abbrev=True,
+                 ui=None):
+        """create a standard commandline parser based on the given class
+        documentation"""
+
+        if ui is None:
+            # mock object:
+            ui = MonkeysignUi(args=[])
+
+        argparse.ArgumentParser.__init__(self, description=ui.__doc__,
+                                         usage=ui.usage,
+                                         epilog=ui.epilog
+                                         % " or ".join(self.configs),
+                                         fromfile_prefix_chars='@',
+                                         formatter_class=argparse.RawDescriptionHelpFormatter)
+        self.add_argument('--version', action='version',
+                            version="\n".join(ui.sysinfo()))
+        self.add_argument('-d', '--debug', action='store_true',
+                            help=_('request debugging information from GPG '
+                                   'engine (lots of garbage)'))
+        self.add_argument('-v', '--verbose', action='store_true',
+                            help=_('explain what we do along the way'))
+        self.add_argument('-n', '--dry-run', dest='dryrun', action='store_true',
+                            help=_('do not actually do anything'))
+        self.add_argument('-u', '--user', nargs='+',
+                            help=_('user id to sign the key with (equivalent '
+                                   'to GPG\'s --local-user option)'))
+        self.add_argument('--certlevel', '--cert-level',
+                            help=_('certification level to sign the key with '
+                                   '(equivalent to GPG\'s '
+                                   '--default-cert-level)'))
+        self.add_argument('-l', '--local', action='store_true',
+                            help=_('import in normal keyring a local '
+                                   'certification'))
+        self.add_argument('-k', '--keyserver',
+                            help=_('keyserver to fetch keys from'))
+        self.add_argument('-s', '--smtpserver', '--smtp',
+                            help=_('SMTP server to use, use a colon to specify '
+                                   'the port number if non-default (%(port)d).'
+                                   ' willl attempt to use STARTTLS to secure '
+                                   'the connexion and fail if unsupported '
+                                   '(default: deliver using the --mta '
+                                   'command)') %
+                            {'port': smtplib.SMTP_PORT})
+        self.add_argument('--tls', action='store_true',
+                            help=_('use a complete TLS connexion instead of '
+                                   'using STARTTLS to upgrade the connexion. '
+                                   'will change the default SMTP port to '
+                                   '%(port)d')
+                            % {'port': smtplib.SMTP_SSL_PORT})
+        self.add_argument('--smtpuser',
+                            help=_('username for the SMTP server '
+                                   '(default: no user)'))
+        self.add_argument('--smtppass',
+                            help=_('password for the SMTP server '
+                                   '(default: prompted, if --smtpuser is '
+                                   'specified)'))
+        self.add_argument('--mta', default="/usr/sbin/sendmail -t",
+                            help=_('command to use to send mail, recipient is '
+                                   'passed on the commandline in the "%%(to)s"'
+                                   ' field, or the command must parse the '
+                                   '"To:" header (default: %(default)s)'))
+        default_mua = "xdg-email --utf8 --to '%%(to)s' --subject '%%(subject)s' --body '%%(body)s' --attach '%%(attach)s'"
+        self.add_argument('--mua', nargs='?', default=None, const=default_mua,
+                            help=_('Mail User Agent to use to send mail. all '
+                                   'parameters are passed on the commandline,'
+                                   'overrides --mta. '
+                                   '(default: %s when specified)')
+                            % default_mua)
+        self.add_argument('--nomail', '--no-mail', action='store_true',
+                            help=_('do not send email at all'))
+        self.add_argument('-t', '--to',
+                            help=_('override destination email for testing '
+                                   '(default: send individually encrypted '
+                                   'email to each uid chosen)'))
+        self.add_argument('--save', action='store_true',
+                          help=_('save the given commandline parameters to the config file'))
+        if 'cli' in ui.__class__.__name__.lower():
+            nargs = '+'
+        else:
+            nargs = '*'
+        self.add_argument('pattern', nargs=nargs,
+                            help=_('pattern of the key to sign (fingerprint,'
+                                   ' user id, etc)'))
+
+    def convert_arg_line_to_args(self, arg_line):
+        """parse a config file"""
+        # skip whitespace and commented lines
+        if re.match('^(#|[\s]*$)', arg_line):
+            return []
+        else:
+            # all lines are assumed to be options
+            # this is important otherwise an attacker could inject key
+            # material in the config file and silently have his key be
+            # signed
+            return ['--' + arg_line]
+
+    def parse_args(self, args=None, namespace=None):
+        """override builtin argument parsing to also parse config files"""
+        configs = map(lambda x: os.path.expanduser(x), self.configs)
+        for conf in configs:
+            try:
+                with open(conf, 'r'):
+                    if args:
+                        args.insert(0, '@' + conf)
+                    else:
+                        sys.argv.insert(1, '@' + conf)
+            except IOError:
+                pass
+        args = argparse.ArgumentParser.parse_args(self,
+                                                  args=args,
+                                                  namespace=namespace)
+        # XXX: a bit clunky because the cli expects this to be the
+        # output of parse_args() while the GTK ui expects this to be
+        # populated as a string, later
+        if len(args.pattern) < 1:
+            args.pattern = None
+        else:
+            # accept space-separated fingerprints
+            args.pattern = "".join(args.pattern)
+        # accept space-separated fingerprints for user as well
+        if args.user is not None:
+            args.user = "".join(args.user)
+        if args.save:
+            self.save_config(args)
+        return args
+
+    def save_config(self, args, config=None, close=True):
+        """save given namespace to default configuration file
+
+        config is a filehandle that will be written to. if None, the
+        first element in the configs array will be opened and written
+        to.
+
+        the config file will not be written to if it doesn't contain
+        the magic header
+
+        performs no exception handling. returns True if all is well
+
+        the config file will be closed unless close=False (used for
+        unit tests)
+        """
+        lines = self.args_to_config(args)
+        if config is None:
+            config = open(os.path.expanduser(self.configs[0]),
+                          'w+')
+        contents = config.read()
+        if len(contents) and self.header not in contents:
+            if close:
+                config.close()
+            return False
+        config.truncate(0)
+        self.write_config(lines, config)
+        if close:
+            config.close()
+        # self.ui.log() except that we don't have logging yet
+        return True
+
+    def args_to_config(self, args):
+        """return given namespace as a configuration file"""
+        lines = []
+        for action in self._actions:
+            if action.dest in self.ephemeral:
+                continue
+            val = vars(args)[action.dest]
+            if val != action.default:
+                if isinstance(action, argparse._StoreAction):
+                    lines.append('%s=%s' % (action.dest, val))
+                else:
+                    lines.append(action.dest)
+        return lines
+
+    def write_config(self, config, fileobj):
+        """take the list of lines and write it to the config file
+
+        we implement our own here because:
+
+        1. ConfigParser is not API stable across Py2/3
+        2. ConfigParser is unnecessarily hard to use and complex
+        3. ConfigParser requires sections which we don't need
+        4. alternatives like ConfigObj are not in the standard library
+        5. this whole config file writer is about 30 lines long
+        """
+        return fileobj.write(self.header + "\n" + "\n".join(config) + "\n")
+
 
 class MonkeysignUi(object):
     """User interface abstraction for monkeysign.
@@ -55,60 +268,40 @@ class MonkeysignUi(object):
 
     You should have a docstring in derived classes, as it will be
     added to the 'usage' output.
-
-    You should also set the usage and epilog parameters, see
-    parse_args().
     """
 
     # what gets presented to the user in the usage (first and last lines)
     # default is to use the OptionParser's defaults
     # the 'docstring' above is the long description
-    usage=None
-    epilog=None
+    # XXX: deprecated, argparse is flexible enough that we do not need
+    # this anymore
+    usage = None
+    epilog = """an arbitrary configuration file may also be supplied on the
+commandline with a @ prefix (e.g. @foo.conf). the following
+configuration files are parsed by default: %s"""
 
     @classmethod
-    def parser(self):
-        """parse the commandline arguments"""
-        parser = optparse.OptionParser(description=self.__doc__, usage=self.usage, epilog=self.epilog, formatter=NowrapHelpFormatter())
-        parser.add_option('--version', dest='version', default=False, action='store_true',
-                          help=_('show version information and quit'))
-        parser.add_option('-d', '--debug', dest='debug', default=False, action='store_true',
-                          help=_('request debugging information from GPG engine (lots of garbage)'))
-        parser.add_option('-v', '--verbose', dest='verbose', default=False, action='store_true',
-                          help=_('explain what we do along the way'))
-        parser.add_option('-n', '--dry-run', dest='dryrun', default=False, action='store_true',
-                          help=_('do not actually do anything'))
-        parser.add_option('-u', '--user', dest='user',
-                          help=_('user id to sign the key with (equivalent to GPG\'s --local-user option)'))
-        parser.add_option('--cert-level', dest='certlevel',
-                          help=_('certification level to sign the key with (equivalent to GPG\'s --default-cert-level)'))
-        parser.add_option('-l', '--local', dest='local', default=False, action='store_true',
-                          help=_('import in normal keyring a local certification'))
-        parser.add_option('-k', '--keyserver', dest='keyserver',
-                          help=_('keyserver to fetch keys from'))
-        parser.add_option('-s', '--smtp', dest='smtpserver', help=_('SMTP server to use, use a colon to specify the port number if non-standard'))
-        parser.add_option('--smtpuser', dest='smtpuser', help=_('username for the SMTP server (default: no user)'))
-        parser.add_option('--smtppass', dest='smtppass', help=_('password for the SMTP server (default: prompted, if --smtpuser is specified)'))
-        parser.add_option('--no-mail', dest='nomail', default=False, action='store_true',
-                          help=_('do not send email at all (default: use sendmail)'))
-        parser.add_option('-t', '--to', dest='to', 
-                          help=_('override destination email for testing (default: send individually encrypted email to each uid chosen)'))
-        return parser
-
-    def parse_args(self, args):
-        parser = self.parser()
-        (self.options, self.pattern) = parser.parse_args(args=args)
-
-        # XXX: a bit clunky because the cli expects this to be the
-        # output of parse_args() while the GTK ui expects this to be
-        # populated as a string, later
-        if len(self.pattern) < 1:
-            self.pattern = None
+    def sysinfo(cls):
+        info = []
+        info.append("Monkeysign: %s" % monkeysign.__version__)
+        info.append('Load path: %s'
+                    % os.path.realpath(os.path.dirname(__file__)))
+        info.append("%s: %s (%s %s)"
+                    % (platform.python_implementation(),
+                       platform.python_version(),
+                       platform.python_compiler(),
+                       " ".join(platform.python_build())))
+        info.append("Kernel: %s" % " ".join(platform.uname()))
+        if 'linux' in platform.system().lower():
+            distribution = " ".join(platform.linux_distribution()).rstrip()
         else:
-            # accept space-separated fingerprints
-            self.pattern = "".join(self.pattern)
-        # make sure parser can be accessed outside of this function
-        return parser
+            distribution = ''
+        info.append("Operating system: %s (%s)"
+                    % (distribution, platform.system()))
+        info.append('PID: %d, CWD: %s' % (os.getpid(), os.getcwd()))
+        info.append('Command: %r' % sys.argv)
+        info.append('GnuPG: %s' % Context().version())
+        return info
 
     def __init__(self, args = None):
         # the options that determine how we operate, from the parse_args()
@@ -131,7 +324,9 @@ class MonkeysignUi(object):
         # temporary, to keep track of the OpenPGPkey we are signing
         self.signing_key = None
 
-        self.parse_args(args)
+        self.parser = MonkeysignArgumentParser(ui=self)
+        self.options = self.parser.parse_args(args=args)
+        self.pattern = self.options.pattern
 
         # set a default logging mechanism
         self.logfile = sys.stderr
@@ -155,9 +350,6 @@ class MonkeysignUi(object):
         # initialize the temporary keyring directory
         self.tmpkeyring = TempKeyring()
 
-        if self.options.version:
-            print(monkeysign.__version__)
-            sys.exit()
         if self.options.debug:
             self.tmpkeyring.context.debug = self.logfile
             self.keyring.context.debug = self.logfile
@@ -177,8 +369,8 @@ class MonkeysignUi(object):
             # no such file or directory is alright: it means the use
             # has no gpg.conf (because we are certain the temp homedir
             # exists at this point)
-            if e.errno != 2:
-                pass
+            if e.errno != errno.ENOENT:
+                raise
 
     def main(self):
         """
@@ -211,12 +403,15 @@ this should not interrupt the flow of the program, but must be visible to the us
 
     def log(self, message):
         """log an informational message if verbose"""
-        if self.options.verbose: print >>self.logfile, message
+        if self.options.verbose or self.options.debug: print >>self.logfile, message
 
     def yes_no(self, prompt, default = True):
         """default UI is not interactive, so we assume yes all the time"""
         return True
 
+    def acknowledge(self, prompt):
+        return True
+
     def choose_uid(self, prompt, uids):
         raise NotImplementedError('choosing not implemented in base class')
 
@@ -249,33 +444,34 @@ this should not interrupt the flow of the program, but must be visible to the us
 we use --secret-keyring instead of copying the secret key material,
 but we still need the public part in the temporary keyring for this to
 work.
-"""
-        self.log(_('copying your private key to temporary keyring in %s') % self.tmpkeyring.homedir)
+
+we copy all keys because we do not want to guess which keys gpg will
+chose. it could vary based on default-key, for example, or some weird
+ordering.
+        """
+        self.log(_('copying your public key to temporary keyring in %s') % self.tmpkeyring.homedir)
         # detect the proper uid
-        if self.options.user is None:
-            keys = self.keyring.get_keys(None, True, False)
-        else:
-            keys = self.keyring.get_keys(self.options.user, True, False)
+        keys = self.keyring.get_keys(self.options.user, True, False)
 
         for fpr, key in keys.iteritems():
             self.log(_('found secret key: %s') % key)
             if not key.invalid and not key.disabled and not key.expired and not key.revoked:
                 self.signing_key = key
-                break
+                # export public key material associated with all private keys
+                # XXX: we should only do export-minimal here, but passing options down is a pain.
+                if not self.tmpkeyring.import_data(self.keyring.export_data(key.fpr)):
+                    self.warning(_('could not export public key material of private key %s') % key.fpr)
 
         if self.signing_key is None:
-            self.abort(_('no default secret key found, abort!'))
-        self.log(_('signing key chosen: %s') % self.signing_key.fpr)
-
-        # export public key material associated with detected private
-        if not self.tmpkeyring.import_data(self.keyring.export_data(self.signing_key.fpr)):
-            self.abort(_('could not find public key material, do you have an OpenPGP key?'))
+            self.abort(_('could not find public key material for any private key, do you have an OpenPGP key?'))
 
     def sign_key(self):
         """sign the key uids, as specified"""
 
         keys = self.tmpkeyring.get_keys(self.pattern)
 
+        # this shouldn't happen unless caller forgot to call copy_secrets
+        assert(keys is not None)  # nosec
         self.log(_('found %d keys matching your request') % len(keys))
 
         for key in keys:
@@ -322,7 +518,8 @@ Sign all identities? [y/N] \
         else:
             from_user = self.signing_key.uidslist[0].uid
 
-        if len(self.signed_keys) < 1: self.warn(_('no key signed, nothing to export'))
+        if len(self.signed_keys) < 1:
+            self.warn(_('no key signed, nothing to export'))
         
         for fpr, key in self.signed_keys.items():
             if self.chosen_uid is None:
@@ -332,14 +529,14 @@ Sign all identities? [y/N] \
                     except GpgRuntimeError as e:
                         self.warn(_('failed to create email: %s') % e)
                         break
-                    self.sendmail(msg)
+                    return self.sendmail(msg)
             else:
                 try:
                     msg = EmailFactory(self.tmpkeyring.export_data(fpr), fpr, self.chosen_uid, from_user, self.options.to)
                 except GpgRuntimeError as e:
                     self.warn(_('failed to create email: %s') % e)
                     break
-                self.sendmail(msg)
+                return self.sendmail(msg)
 
     def sendmail(self, msg):
             """actually send the email
@@ -347,7 +544,10 @@ Sign all identities? [y/N] \
 expects an EmailFactory email, but will not mail if nomail is set"""
             if self.options.smtpserver is not None and not self.options.nomail:
                 if self.options.dryrun: return True
-                server = smtplib.SMTP()
+                if self.options.tls:
+                    server = smtplib.SMTP_SSL()
+                else:
+                    server = smtplib.SMTP()
                 server.set_debuglevel(self.options.debug)
                 # to be nicer to users, we could catch socket.error exceptions from
                 # server.connect() here and display a meaningful message to stderr.
@@ -359,9 +559,10 @@ expects an EmailFactory email, but will not mail if nomail is set"""
                     self.abort(_('Unexpected SMTP server error while talking to %s, code: %s (%s)') % (self.options.smtpserver, code, srvmsg))
                 try:
                     server.starttls()
-                except SMTPException:
+                except smtplib.SMTPException:
                     self.warn(_('SMTP server does not support STARTTLS'))
-                    if self.options.smtpuser: self.warn(_('authentication credentials will be sent in clear text'))
+                    if self.options.smtpuser:
+                        self.abort(_('aborting authentication as credentials would have been sent in clear text'))
                 if self.options.smtpuser:
                     if not self.options.smtppass:
                         self.options.smtppass = self.prompt_pass(_('enter SMTP password for server %s: ') % self.options.smtpserver)
@@ -372,15 +573,66 @@ expects an EmailFactory email, but will not mail if nomail is set"""
                 return True
             elif not self.options.nomail:
                 if self.options.dryrun: return True
-                p = subprocess.Popen(['/usr/sbin/sendmail', '-t'], stdin=subprocess.PIPE)
-                p.communicate(msg.as_string().encode('utf-8'))
-                self.warn(_('sent message through sendmail to %s') % msg.mailto)
+                if self.options.mua:
+                    keyfile = tempfile.NamedTemporaryFile()
+                    # XXX: this should be part of the EmailFactory
+                    keymaterial = msg.tmpkeyring.export_data(msg.keyfpr)
+                    encrypted = msg.tmpkeyring.encrypt_data(keymaterial,
+                                                            msg.keyfpr)
+                    keyfile.write(encrypted)
+                    keyfile.flush()
+                try:
+                    if self.options.mua:
+                        command = [x % {'to': msg.mailto,
+                                        'body': msg.body,
+                                        'subject': msg.subject,
+                                        'attach': keyfile.name}
+                                   for x in shlex.split(self.options.mua)]
+                    else:
+                        command = [x % {'to': msg.mailto}
+                                   for x in shlex.split(self.options.mta)]
+                except ValueError as e:
+                    if 'incomplete format' in str(e):
+                        self.warn(_('invalid command (%s) specified to send email: %s')
+                                  % (self.options.mua or self.options.mta, str(e)))
+                        return False
+                    else:
+                        raise
+                self.log('running command %s' % command)
+                try:
+                    p = subprocess.Popen(command,
+                                         stdin=subprocess.PIPE,
+                                         stdout=subprocess.PIPE,
+                                         stderr=subprocess.PIPE)  # nosec
+                except OSError as e:
+                    self.warn(_('cannot find MTA %s, try specifying --mua, --mta or --smtp: %s')
+                              % (command, e))
+                    return False
+
+                if self.options.mua:
+                    tosend = None
+                else:
+                    tosend = msg.as_string().encode('utf-8')
+                self.mailer_output, self.mailer_error = p.communicate(tosend)
+                if self.options.mua:
+                    self.acknowledge('when you have finished writing the email')
+                    keyfile.close()  # remove the tmpfile
+                if p.returncode == 0:
+                    self.warn(_('sent message to %(destination)s with %(command)s')
+                              % {'destination': msg.mailto,
+                                 'command': self.options.mua or self.options.mta})
+                else:
+                    self.warn(_('failed sending message to %(destination)s with %(command)s: %(error)s')
+                              % {'destination': msg.mailto,
+                                 'command': self.options.mua or self.options.mta,
+                                 'error': self.mailer_output + self.mailer_error})
+                return p.returncode == 0
             else:
                 # okay, no mail, just dump the exported key then
                 self.warn(_("""\
-not sending email to %s, as requested, here's the email message:
+not sending email to %s, as requested, here's the encrypted signed public key you can paste in your email client:
 
-%s""") % (msg.mailto, msg.create_mail_from_block()))
+%s""") % (msg.mailto, msg.encrypted()))
 
 
 class EmailFactory:
@@ -459,12 +711,25 @@ mailto: who to send the mail to (usually similar to recipient, but can be used t
             for uid in todelete:
                 self.tmpkeyring.del_uid(fpr, uid)
 
+    def encrypted(self):
+        '''encrypted blob of the MIME multipart message
+
+        this is designed to be attached as the main body of an email,
+        but can also be copy-pasted anywhere, although the decrypted
+        version will look a little ackward because it will include all
+        sorts of MIME stuff
+
+        but if proper headers are added to the message, it should be
+        parsed okay by OpenPGP and MIME-capable clients.'''
+
+        msg = self.create_mail_from_block().as_string()
+        return self.tmpkeyring.encrypt_data(msg, self.keyfpr)
+
     def get_message(self):
         # first layer, seen from within:
         # an encrypted MIME message, made of two parts: the
         # introduction and the signed key material
-        message = self.create_mail_from_block()
-        encrypted = self.tmpkeyring.encrypt_data(message.as_string(), self.keyfpr)
+        encrypted = self.encrypted()
 
         # the second layer up, made of two parts: a version number
         # and the first layer, encrypted
@@ -514,9 +779,3 @@ mailto: who to send the mail to (usually similar to recipient, but can be used t
         name, address = parseaddr(self.mailto)
         msg['To'] = formataddr((Header(name.encode('utf-8'), 'UTF-8').encode(), address))
         return msg
-
-class NowrapHelpFormatter(optparse.IndentedHelpFormatter):
-    """A non-wrapping formatter for OptionParse."""
-
-    def _format_text(self, text):
-        return text
