@@ -21,7 +21,7 @@ import sys
 import os
 import time
 
-sys.path.append(os.path.dirname(__file__))
+sys.path.insert(0, os.path.dirname(__file__))
 
 suite = unittest.TestLoader().discover('tests')
 unittest.TextTestRunner().run(suite)
