@@ -27,7 +27,7 @@ import unittest
 import tempfile
 import re
 
-sys.path.append(os.path.dirname(__file__) + '/..')
+sys.path.insert(0, os.path.dirname(__file__) + '/..')
 
 from monkeysign.gpg import *
 
@@ -120,7 +120,7 @@ class TestKeyringBase(unittest.TestCase):
 
     def tearDown(self):
         """trash the temporary directory we created"""
-        shutil.rmtree(self.tmp)
+        shutil.rmtree(self.tmp, ignore_errors=True)
 
 class TestKeyringBasics(TestKeyringBase):
     """Test the Keyring class base functionality."""
@@ -186,23 +186,31 @@ class TestKeyringBasics(TestKeyringBase):
         however, gpg returns the wrong exit code here, so we end up at
         looking if there is really no output
         """
-        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) + '/96F47C6A-secret.asc').read()))
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__)
+                                                  + '/96F47C6A-secret.asc').read()))
         with self.assertRaises(GpgRuntimeError):
             self.gpg.sign_key('7B75921E')
 
     def test_failed_revoke(self):
-        self.gpg.import_data(open(os.path.dirname(__file__) + '/96F47C6A.asc').read())
-        self.gpg.import_data(open(os.path.dirname(__file__) + '/96F47C6A-revoke.asc').read())
-        self.gpg.import_data(open(os.path.dirname(__file__) + '/7B75921E.asc').read())
+        self.gpg.import_data(open(os.path.dirname(__file__)
+                                  + '/96F47C6A.asc').read())
+        self.gpg.import_data(open(os.path.dirname(__file__)
+                                  + '/96F47C6A-revoke.asc').read())
+        self.gpg.import_data(open(os.path.dirname(__file__)
+                                  + '/7B75921E.asc').read())
         with self.assertRaises(GpgRuntimeError):
             self.gpg.sign_key('7B75921E', True)
 
+
 class TestKeyringWithKeys(TestKeyringBase):
     def setUp(self):
         TestKeyringBase.setUp(self)
-        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) + '/7B75921E.asc').read()))
-        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) + '/96F47C6A.asc').read()))
-        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) + '/96F47C6A-secret.asc').read()))
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__)
+                                                  + '/7B75921E.asc').read()))
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__)
+                                                  + '/96F47C6A.asc').read()))
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__)
+                                                  + '/96F47C6A-secret.asc').read()))
 
     def test_get_keys(self):
         """test that we can list the keys after importing them
@@ -248,8 +256,9 @@ class TestKeyringWithKeys(TestKeyringBase):
     def test_sign_already_signed(self):
         """test if signing a already signed key fails with a meaningful message"""
         self.assertTrue(self.gpg.sign_key('Antoine Beaupré <anarcat@debian.org>'))
-        with self.assertRaises(GpgRuntimeError):
+        with self.assertRaises(GpgRuntimeError) as e:
             self.gpg.sign_key('Antoine Beaupré <anarcat@debian.org>')
+            self.assertIn('you already signed that key', str(e))
 
     def test_encrypt_decrypt_data_armored_untrusted(self):
         """test if we can encrypt data to our private key (and decrypt it)"""
@@ -324,6 +333,24 @@ class TestKeyringWithKeys(TestKeyringBase):
         """test verify_file()"""
         self.assertTrue(self.gpg.verify_file(os.path.dirname(__file__) + '/testfile.txt.asc', os.path.dirname(__file__) + '/testfile.txt'))
 
+    def test_sign_revoked_uid(self):
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__)
+                                                  + '/323F39BD.asc').read()))
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__)
+                                                  + '/323F39BD-secret.asc').read()))
+        self.gpg.import_data(open(os.path.dirname(__file__)
+                                  + '/96F47C6A-revuid.asc').read())
+        # designate that new key as the signing key
+        self.gpg.context.set_option('local-user', '323F39BD')
+        self.assertTrue(self.gpg.sign_key('3F94240C918E63590B04152E86E4E70A96F47C6A', True))
+
+        self.gpg.context.call_command(['list-sigs', '96F47C6A'])
+        # try to find a revoked uid that is signed
+        for uid in self.gpg.context.stdout.split('uid:'):
+            if 'rev:' in uid:
+                self.assertNotIn('sig:::1:A31E75E4323F39BD', uid)
+
+
 class TestKeyringWithAbnormalKeys(TestKeyringBase):
     """this tests specifically weird keys"""
     def setUp(self):
@@ -332,16 +359,47 @@ class TestKeyringWithAbnormalKeys(TestKeyringBase):
         self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) + '/96F47C6A-secret.asc').read()))
 
     def test_wrongly_place_sigs(self):
-        """this provokes an exception with:
+        """test zack's key
 
-    monkeysign.gpg.GpgProtocolError: [Errno 0] expected "^\[GNUPG:\] GOT_IT", found "gpg: moving a key signature to the correct place"
+        this provokes an exception with:
 
-seems like it's a key not respecting standard: http://lists.nongnu.org/archive/html/sks-devel/2012-07/msg00122.html"""
-        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) + '/6D866396.asc').read()))
+        monkeysign.gpg.GpgProtocolError: [Errno 0] expected
+        "^\[GNUPG:\] GOT_IT", found "gpg: moving a key signature to
+        the correct place"
+
+        seems like it's a key not respecting standard:
+        http://lists.nongnu.org/archive/html/sks-devel/2012-07/msg00122.html
+
+        see https://bugs.debian.org/736120
+        """
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) +
+                                                  '/6D866396.asc').read()))
         self.assertTrue(self.gpg.sign_key('6D866396', True))
         self.gpg.context.call_command(['list-sigs', '6D866396'])
         self.assertRegexpMatches(self.gpg.context.stdout, 'sig:::1:86E4E70A96F47C6A:[^:]*::::Second Test Key <unittests@monkeysphere.info>:10x:')
 
+    def test_broken_encoding(self):
+        """test some key that has a non-standard encoding
+
+        RFC4880 specifies that UIDs should be UTF-8, but someone this
+        one isn't.
+
+        see https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=736629
+        """
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) +
+                                                  '/ECAA37C45C7E48CE.asc').read()))
+        key = self.gpg.get_keys('095D9EC8C995AB203DC260FEECAA37C45C7E48CE')
+        key['095D9EC8C995AB203DC260FEECAA37C45C7E48CE'].__str__().encode('utf-8')
+
+    def test_expired_subkeys(self):
+        """test a key that has an expired subkey"""
+        self.assertTrue(self.gpg.import_data(open(os.path.dirname(__file__) +
+                                                  '/576407A629299233.asc').read()))
+        self.assertTrue(self.gpg.sign_key('576407A629299233', True))
+        self.gpg.context.call_command(['list-sigs', '576407A629299233'])
+        self.assertRegexpMatches(self.gpg.context.stdout, 'sig:::1:86E4E70A96F47C6A:[^:]*::::Second Test Key <unittests@monkeysphere.info>:10x:')
+
+
 class TestOpenPGPkey(unittest.TestCase):
     def setUp(self):
         self.key = OpenPGPkey("""tru::1:1343350431:0:3:1:5
